#include <filein/rofiworld_loader.hpp>

#include <algo/misc.hpp>
#include <filein/json_loader.hpp>

void export_rofiworld(const rofiworld_ptr rofiworld, const std::vector<objectbase_ptr> &selected_spaceJoints, const char* file_name)
{
    boost::json::value value;
    auto rofiworld_object = value.emplace_object();
    
    boost::json::array modules_array;
    export_modules(modules_array, rofiworld->getModulesMap());
    rofiworld_object.emplace("modules", modules_array);


    boost::json::array spaceJoints_array;
    export_spaceJoints(spaceJoints_array, rofiworld, selected_spaceJoints);
    rofiworld_object.emplace("spaceJoints", spaceJoints_array);

    boost::json::array moduleJoints_array;
    export_moduleJoints(moduleJoints_array, rofiworld);
    rofiworld_object.emplace("moduleJoints", moduleJoints_array);

    save_configuration(rofiworld_object, "./data/rofi/rofiworlds", file_name);
}

void export_modules(boost::json::array &modules_array, const std::map<module_ptr, uint64_t> &modules_map)
{
    for (const auto &[module, id] : modules_map)
    {
        const auto &module_type = module->getType();

        if (module_type == "cube")
            export_module_cube(modules_array, module, id); // TO DO
        if (module_type == "universal")
            export_module_universal(modules_array, module, id);
        if (module_type == "pad")
            export_module_pad(modules_array, module, id);
        if (module_type == "unknown")
            export_module_unknown(modules_array, module, id); // TO DO
    }
}

void export_module_cube(boost::json::array &modules_array, const module_ptr module, const uint64_t &id)
{
    // TO DO
}
void export_module_universal(boost::json::array &modules_array, const module_ptr module, const uint64_t &id)
{
    boost::json::object module_object;

    export_alpha_beta(module_object, module);

    export_gamma(module_object, module);

    module_object.emplace("id", id);
    module_object.emplace("type", "universal");

    modules_array.emplace_back(module_object);
}

void export_alpha_beta(boost::json::object &module_object, const module_ptr module)
{
    using namespace rofi;
    float alpha = 0;
    float beta = 0;

    for (const auto&[connector, dock] : module->getConnectorDocks())
    {
        if (dock == "A-Z")
            alpha = std::dynamic_pointer_cast<RotationConnector>(connector)->getCurrentRotation();
        else if (dock == "B-Z")
            beta = std::dynamic_pointer_cast<RotationConnector>(connector)->getCurrentRotation();
    }

    module_object.emplace("alpha", alpha);
    module_object.emplace("beta", beta);
}

void export_gamma(boost::json::object &module_object, const module_ptr module)
{
    const auto fst_orientation = module->getParts()->getVoxels().front()->getNode()->getRotationAxisYWorld();
    const auto snd_orientation = module->getParts()->getVoxels().back()->getNode()->getRotationAxisYWorld();
    const auto ref_axis = module->getParts()->getVoxels().front()->getNode()->getRotationAxisZWorld();
    auto gamma = glm::degrees(glm::orientedAngle(fst_orientation, snd_orientation, ref_axis));
    module_object.emplace("gamma", gamma);
}

void export_module_pad(boost::json::array &modules_array, const module_ptr module, const uint64_t &id)
{
    boost::json::object module_object;

    const auto [width, height] = get_pad_width_height_from_connector_count(module->getParts()->getConnections());

    module_object.emplace("width", width);
    module_object.emplace("height", height);

    module_object.emplace("id", id);
    module_object.emplace("type", "pad");

    modules_array.emplace_back(module_object);
}

void export_module_unknown(boost::json::array &modules_array, const module_ptr module, const uint64_t &id)
{
    // TO DO
}

void export_spaceJoints(boost::json::array &spaceJoints_array, const rofiworld_ptr rofiworld, const std::vector<objectbase_ptr> &selected_space_joints)
{
    using namespace rofi;

    for (const auto &component : selected_space_joints)
    {
        if (const auto voxel = std::dynamic_pointer_cast<Voxel>(component))
        {
            auto voxel_node = voxel->getNode();
            auto voxel_id = rofiworld->getModuleWithNode(voxel_node)->getComponentID(voxel);
            export_spaceJoint(spaceJoints_array, rofiworld, voxel_node, voxel_id);
        }
        else if (const auto connector = std::dynamic_pointer_cast<Connector>(component))
        {
            auto connector_node = connector->getNode();
            auto connector_id = rofiworld->getModuleWithNode(connector_node)->getComponentID(connector);
            export_spaceJoint(spaceJoints_array, rofiworld, connector_node, connector_id);
        }
    }
}

void export_spaceJoint(boost::json::array &spaceJoints_array, const rofiworld_ptr rofiworld, const node_ptr node, const uint64_t component_id)
{
    boost::json::object spaceJoint_object;

    export_spaceJoint_point(spaceJoint_object, node);
    export_spaceJoint_joint(spaceJoint_object, node);
    export_spaceJoint_to(spaceJoint_object, rofiworld, rofiworld->getModulesMap().at(rofiworld->getModuleWithNode(node)), component_id);

    spaceJoints_array.emplace_back(spaceJoint_object);
}

void export_spaceJoint_point(boost::json::object &spaceJoint_object, const node_ptr node)
{
    boost::json::array point_array;
    const auto position = convertToRoFICoordinates(node->getPositionWorld());
    // const auto position = node->getPositionWorld();

    for (int i = 0; i < 3; ++i)
        point_array.emplace_back(position[i]);

    spaceJoint_object.emplace("point", point_array);
}

void export_spaceJoint_joint(boost::json::object &spaceJoint_object, const node_ptr node)
{
    boost::json::object joint_object;

    joint_object.emplace("type", "rigid");

    export_spaceJoint_joint_matrix(joint_object, node);

    boost::json::array positions_array;
    joint_object.emplace("positions", positions_array);

    spaceJoint_object.emplace("joint", joint_object);
}

void export_spaceJoint_joint_matrix(boost::json::object &joint_object, const node_ptr node)
{
    auto world_mat = convertToRoFICoordinates(node->getModelMatWorld());
    // auto world_mat = node->getModelMatWorld();
    float epsilon = 0.00001f;
    bool is_identity = true;

    for (int i = 0; i < 4; ++i)
    {
        auto id_column = glm::vec4(0.0f);
        id_column[i] = 1.0f;
        if (!glm::all(glm::epsilonEqual(world_mat[i], id_column, epsilon)))
            is_identity = false;
    }
    if (is_identity)
    {
        joint_object.emplace("sourceToDestination", "identity");
        return;
    }

    boost::json::array sourceToDestination_array;

    for (int i = 0; i < 4; ++i)
    {
        boost::json::array row_array;
        row_array.emplace_back(glm::row(world_mat, i).x);
        row_array.emplace_back(glm::row(world_mat, i).y);
        row_array.emplace_back(glm::row(world_mat, i).z);
        row_array.emplace_back(glm::row(world_mat, i).w);
        sourceToDestination_array.emplace_back(row_array);
    }

    joint_object.emplace("sourceToDestination", sourceToDestination_array);
}

void export_spaceJoint_to(boost::json::object &spaceJoint_object, const rofiworld_ptr rofiworld, const uint64_t module_id, uint64_t component_id)
{
    boost::json::object to_object;

    to_object.emplace("component", component_id);
    to_object.emplace("id", module_id);

    spaceJoint_object.emplace("to", to_object);
}

void export_moduleJoints(boost::json::array &moduleJoints_array, const rofiworld_ptr rofiworld)
{
    using namespace rofi;
    for (const auto &voxel : rofiworld->getVoxelGraph()->getVoxels())
    {
        const auto module_link = std::dynamic_pointer_cast<ModuleLink>(voxel);
        if (!module_link)
            continue;

        export_moduleJoint(moduleJoints_array, rofiworld, module_link);
    }
}

void export_moduleJoint(boost::json::array &moduleJoints_array, const rofiworld_ptr rofiworld, const std::shared_ptr<rofi::ModuleLink> module_link)
{
    boost::json::object moduleJoint_object;

    const auto from_con = module_link->getInterModuleConnectors().first.lock();
    const auto to_con = module_link->getInterModuleConnectors().second.lock();

    const auto from_module = rofiworld->getModuleWithNode(from_con->getNode());
    const auto to_module = rofiworld->getModuleWithNode(to_con->getNode());

    export_moduleJoint_fromto(moduleJoint_object, from_module, rofiworld->getModulesMap().at(from_module), from_con, "from");
    export_moduleJoint_fromto(moduleJoint_object, to_module, rofiworld->getModulesMap().at(to_module), to_con, "to");

    export_moduleJoint_orientation(moduleJoint_object, from_con, to_con);

    moduleJoints_array.emplace_back(moduleJoint_object);
}

void export_moduleJoint_fromto(boost::json::object &moduleJoint_object, const module_ptr module, const uint64_t module_id, const connector_ptr connector, const std::string &&where)
{
    boost::json::object from_to_object;

    from_to_object.emplace("id", module_id);

    if (module->getType() == "universal")
        from_to_object.emplace("connector", module->getConnectorDock(connector));
    else
        from_to_object.emplace("connector", module->getComponentID(connector));

    moduleJoint_object.emplace(where, from_to_object);
}

void export_moduleJoint_orientation(boost::json::object &moduleJoint_object, const connector_ptr from_con, const connector_ptr to_con)
{
    using namespace rofi;
    using enum cardinal;

    std::string orientation;

    switch(Connector::get_mutual_orientation(from_con, to_con))
    {
        case North: 
            orientation = "North";
            break;
        case East:  
            orientation = "East";
            break;
        case South:  
            orientation = "South";
            break;
        case West:  
            orientation = "West";
            break;
        default:
            ASSUMPTION(false);
    }

    moduleJoint_object.emplace("orientation", orientation);
}

std::vector<rofiworld_ptr> import_saved_configurations()
{
    const std::filesystem::path rofiworlds_directory = "./data/rofi/rofiworlds";
    std::vector<rofiworld_ptr> imported_configurations;

    for (auto const& dir_entry : std::filesystem::directory_iterator{rofiworlds_directory})
    {
        if (dir_entry.path().filename().string() == "." || dir_entry.path().filename().string() == "..")
            continue;

        const auto rofiworld = import_rofiworld(dir_entry);
        if (rofiworld)
            imported_configurations.emplace_back(rofiworld);
    }

    return imported_configurations;
}

rofiworld_ptr import_rofiworld(const std::filesystem::path file_path)
{
    boost::json::array modules_array;
    boost::json::array spaceJoints_array;
    boost::json::array moduleJoints_array;

    get_rofiworld_data(modules_array, spaceJoints_array, moduleJoints_array);
}

void get_rofiworld_data(const boost::json::array &modules_array, const boost::json::array &spaceJoints_array, const boost::json::array &moduleJoints_array)
{
    
}
