#include <rofi/module.hpp>
namespace rofi
{

void Module::makeNodeTree()
{
    for (auto &voxel: parts->getVoxels())
        node->addChild(voxel->getNode());

    for (auto &connector : parts->getConnections())
        node->addChild(connector->getNode());
}

void Module::setConnectorDocks() // TO DO: change if connector placement gets changed
{
    if (type != "universal")
        return;

    using enum side;
    std::string dock;
    auto a_pos = glm::vec3(0.0f);

    for (const auto &connector : parts->getConnections())
    {
        if (connector->getType(false) == connector_type::shared_rotation)
            continue;

        auto con_side = connector->getVoxelSide().first;
        auto voxel_pos = connector->getConnection().first.lock()->getNode()->getPositionLocal();
        switch(con_side)
        {
            case x_neg:
                dock = voxel_pos == a_pos ? "A+X" : "B+X";
                break;
            case x_pos:
                dock = voxel_pos == a_pos ? "A-X" : "B-X";
                break;
            case z_neg:
                dock = voxel_pos == a_pos ? "A-Z" : "B-Z";
                break;
            default:
                ASSUMPTION(false);
        }
        connector_docks.emplace(connector, dock);
    }
}

void Module::setComponentIDs()
{
    for (const auto connector : parts->getConnections())
    {
        if (connector->getType(true) == connector_type::open)
            component_ids.emplace(connector, component_ids.size());
    }
    for (const auto connector : parts->getConnections())
    {
        if (connector->getType(true) == connector_type::rotation)
            component_ids.emplace(connector, component_ids.size());
    }

    for (const auto voxel : parts->getVoxels())
    {
        if (!std::dynamic_pointer_cast<ModuleLink>(voxel) && !std::dynamic_pointer_cast<PadBoard>(voxel))
            component_ids.emplace(voxel, component_ids.size());
    }
}

Module::Module(const Module &other)
{
    node = Node::copy_deep(other.node);
    parts = VoxelGraph::copy(other.parts, node->getChildren());
    name = other.name;
    type = other.type;
    setComponentIDs();
    setConnectorDocks();
}
Module::Module(const Module &&other)
{
    node = std::move(other.node);
    parts = std::move(other.parts);
    name = std::move(other.name);
    type = std::move(other.type);
    setComponentIDs();
    setConnectorDocks();
}
Module& Module::operator=(const Module &other)
{
    node = Node::copy_deep(other.node);
    parts = VoxelGraph::copy(other.parts, node->getChildren());
    name = other.name;
    type = other.type;
    component_ids = other.component_ids;
    setConnectorDocks();

    return *this;
}
Module& Module::operator=(const Module &&other)
{
    node = std::move(other.node);
    parts = std::move(other.parts);
    name = std::move(other.name);
    type = std::move(other.type);
    setComponentIDs();
    setConnectorDocks();

    return *this;
}

Module::Module(voxel_graph_ptr _parts, std::string &_name, std::string &_type, std::map<objectbase_ptr, uint64_t> &_component_ids)
: node{Node::create()}, parts{_parts}, name{_name}, type{_type}, component_ids{_component_ids}
{
    ASSUMPTION(!_parts->getVoxels().empty() || !_parts->getConnections().empty());
    makeNodeTree();
    setConnectorDocks();
}

Module::Module(voxel_graph_ptr _parts, std::string &&_name, std::string &&_type, std::map<objectbase_ptr, uint64_t> &&_component_ids)
: node{Node::create()}, parts{_parts}, name{std::move(_name)}, type{std::move(_type)}, component_ids{std::move(_component_ids)}
{
    ASSUMPTION(!_parts->getVoxels().empty() || !_parts->getConnections().empty());
    makeNodeTree();
    setConnectorDocks();
}

module_ptr Module::copy(const module_ptr other_module)
{
    return module_ptr{ new Module(*other_module) };
}

}
